!
!   LAPACK benchmarking code
!   Author: David Knezevic, October 2006
!



program main
  implicit none

! Benchmarking codes, uncomment a subroutine to time a LAPACK routine
!  call dense_linear_solve_demo
  call tridiag_linear_solve_demo
!  call SVD_demo
!  call EVP_demo
!  call LLS_qr_demo

end


subroutine dense_linear_solve_demo
  ! Solve linear systems for a range of sizes of dense random matrices
  implicit none

  integer, parameter :: MIN_J = 1
  integer, parameter :: MAX_J = 11
  integer N, j, m
  integer info      ! factorisation/solving error flag
  integer seed

  integer,         dimension(:),   allocatable :: ipiv
  real (kind = 8), dimension(:,:), allocatable :: A
  real (kind = 8), dimension(:),   allocatable :: b

  real, parameter :: T = 2.
  real (kind = 8) t0, t1, elapsed

  write(*, * )
  write(*, * ), '-------------------------------------------------------'
  write(*, * )
  write(*, * ), 'LAPACK benchmarking for dense linear systems:'
  write(*, * )
  write(*, * )

  seed = 123456789 ! seed gets updated in the generator, only need to initialise it

  do j=MIN_J, MAX_J

       N = 2.**j
       elapsed = 0.
       m = 0

10     if (elapsed < T) then

         allocate( A(N,N) )
         allocate( b(N) )
         allocate( ipiv(N) )

         call dmat_uniform_01 ( N, N, seed, A )
         call dmat_uniform_01 ( N, 1, seed, b )

         call cpu_time(t0)
         call dgetrf ( N, N, A, N, ipiv, info )            ! LU factorise the matrix
         call dgetrs ( 'n', N, 1, A, N, ipiv, b, N, info ) ! Solve the linear system
         call cpu_time(t1)

         elapsed = elapsed + (t1-t0)

         deallocate( A, b, ipiv )

         m = m + 1
         goto 10
       end if

       write(*, '(a,i2,a,d10.5)'), ' N = 2^', j, ', Average time for linear solve = ', elapsed/m

  end do

  write(*, * )

end

subroutine tridiag_linear_solve_demo
  ! Solve linear systems for a range of sizes of random tridiagonal matrices
  implicit none

  integer, parameter :: MIN_I = 1
  integer, parameter :: MAX_I = 20
  integer N, i, m
  integer info      ! factorisation/solving error flag
  integer seed

  integer,         dimension(:),   allocatable :: ipiv
  real (kind = 8), dimension(:),   allocatable :: b, c, d, e

  real (kind = 8), parameter :: T = 2.
  real (kind = 8) t0, t1, elapsed

  write(*, * ), '-------------------------------------------------------'
  write(*, * )
  write(*, * ), 'LAPACK benchmarking for tridiagonal system:'
  write(*, * )
  write(*, * )

  seed = 123456789 ! seed gets updated in the generator, only need to initialise it


   do i=MIN_I, MAX_I

     N = 2**i
     elapsed = 0.
     m = 0

10   if (elapsed < T) then

       allocate( b(N) )
       allocate( c(N-1) )
       allocate( d(N) )
       allocate( e(N-1) )

       call dmat_uniform_01 ( N,   1, seed, b)
       call dmat_uniform_01 ( N-1, 1, seed, c)
       call dmat_uniform_01 ( N,   1, seed, d)
       call dmat_uniform_01 ( N-1, 1, seed, e)

       call cpu_time(t0)
       call dgtsv ( N, 1, c, d, e, b, N, info )
       call cpu_time(t1)

       elapsed = elapsed + (t1-t0)
       deallocate(b, c, d, e)

       m = m+1

       goto 10
     end if

     write(*, '(a,i2,a,d10.5)'), ' N = 2^', i, ', average time elapsed for tridiagonal linear solve = ', elapsed/m

  end do

  write(*, * )

end


subroutine EVP_demo
!***********************************************************************
! Benchmarking code for general nonsymmetric eigenvalue problem

  implicit none

  integer, parameter :: MIN_I = 1
  integer, parameter :: MAX_I = 9
  integer N, i, j, m, info, seed, lwork

  real (kind = 8), dimension(:,:), allocatable :: A, vl, vr
  real (kind = 8), dimension(:),   allocatable :: wr, wi, work

  real (kind = 8), parameter :: T = 2.
  real (kind = 8) t0, t1, elapsed

  write ( *, '(a)' ) ' -------------------------------------------------------'
  write ( *, '(a)' ) ' '
  write ( *, '(a)' ) '  LAPACK benchmarking for computation of left '
  write ( *, '(a)' ) '  eigenvalues and eigenvectors'
  write ( *, '(a)' ) ' '
  
  seed = 123456789

   do i=MIN_I, MAX_I

     N = 2**i
     lwork = 4*N

     allocate( work(lwork) )
     allocate( vl(N,N) )
     allocate( vr(N,N) )
     allocate( wr(N) )
     allocate( wi(N) )

     elapsed = 0.
     m = 0

10   if (elapsed < T) then

       allocate( A(N,N) )

       call dmat_uniform_01 ( N, N, seed, A )

       call cpu_time(t0)
       call dgeev('v', 'n',N, A, N, wr, wi, vl, N, vr, N, work, lwork, info)
       call cpu_time(t1)

       elapsed = elapsed + (t1-t0)

       deallocate(A)

       m = m+1

       goto 10
     end if

     deallocate(work,vl, vr, wr, wi)

     write(*, '(a,i2,a,d10.5)'), ' N = 2^', i, ', average time elapsed for eigenvalue/vector computation = ', elapsed/m
   end do

   write ( *, '(a)' ) ' '
end


subroutine SVD_demo
!***********************************************************************
!
!! LAPACK SVD demo written modified (by David Knezevic, October 2006), from 
!  Test05 in lapack_prb.f90 available from 
!  http://www.scs.fsu.edu/~burkardt/f_src/lapack/lapack.html 
!  Author: John Burkardt, 12 September 2006
!
!  Discussion:
!
!    This code computes the singular value decomposition:
!
!      A = U * S * V'
!
!    for a range of values of N

  implicit none

  integer, parameter :: MIN_I = 1
  integer, parameter :: MAX_I = 9
  integer N, i, j, m, info, seed, lwork

  integer,         dimension(:),   allocatable :: ipiv
  real (kind = 8), dimension(:,:), allocatable :: a, sigma, u, vt
  real (kind = 8), dimension(:),   allocatable :: s, work

  real (kind = 8), parameter :: T = 2.
  real (kind = 8) t0, t1, elapsed

  write ( *, '(a)' ) ' -------------------------------------------------------'
  write ( *, '(a)' ) ' '
  write ( *, '(a)' ) '  LAPACK benchmarking for singular value decomposition'
  write ( *, '(a)' ) '  (all left and right singular vectors computed):'
  write ( *, '(a)' ) ' '
  write ( *, '(a)' ) '    A = U * S * V'''
  write ( *, '(a)' ) ' '
  write ( *, '(a)' ) ' '
  
  seed = 123456789

   do i=MIN_I, MAX_I

     N = 2**i
     lwork = 5*N

     elapsed = 0.
     m = 0

10   if (elapsed < T) then

       allocate( A(N,N) )
       allocate( s(N) )
       allocate( sigma(N,N) )
       allocate( u(N,N) )
       allocate( vt(N,N) )
       allocate( work(lwork) )

       call dmat_uniform_01 ( N, N, seed, A )

       call cpu_time(t0)
       call dgesvd ('A','A',N,N,A,N,s,u,N,vt,N,work,lwork,info)

       sigma(1:N,1:N) = 0.0D+00
       do j = 1, N
         sigma(j,j) = s(j)
       end do

       a(1:N,1:N) = matmul ( u(1:N,1:N), matmul ( sigma(1:N,1:N), vt(1:N,1:N) ) )

       call cpu_time(t1)
       elapsed = elapsed + (t1-t0)

       deallocate(A, s, sigma, u, vt, work)

       m = m+1

       goto 10
     end if

     write(*, '(a,i2,a,d10.5)'), ' N = 2^', i, ', average time elapsed for SVD = ', elapsed/m
   end do

   write ( *, '(a)' ) ' '
end


subroutine LLS_qr_demo
!***********************************************************************
! Benchmarking code for over-determined linear least squares problem

  implicit none

  integer, parameter :: MIN_I = 1
  integer, parameter :: MAX_I = 11
  integer N, i, j, m, info, seed, lwork

  real (kind = 8), dimension(:,:), allocatable :: A
  real (kind = 8), dimension(:),   allocatable :: b, work

  real (kind = 8), parameter :: T = 2.
  real (kind = 8) t0, t1, elapsed

  write ( *, '(a)' ) ' -------------------------------------------------------'
  write ( *, '(a)' ) ' '
  write ( *, '(a)' ) '  LAPACK benchmarking for linear least squares problem '
  write ( *, '(a)' ) '  using QR factorisation where A is a 2NxN matrix:'
  write ( *, '(a)' ) ' '

  seed = 123456789

   do i=MIN_I, MAX_I

     N = 2**i
     lwork = 3*N*N

     allocate( work(lwork) )

     elapsed = 0.
     m = 0

10   if (elapsed < T) then

       allocate( A(2*N,N) )
       allocate( b(2*N) )

       call dmat_uniform_01 ( 2*N, N, seed, A )
       call cpu_time(t0)
       call dgels('N', 2*N, N, 1, A, 2*N, b, 2*N, work, lwork, info)
       call cpu_time(t1)

       if ( info /= 0 ) then
         write ( *, '(a)' ) ' '
         write ( *, '(a,i8)' ) '  Solution procedure failed, INFO = ', info
         return
       end if

       elapsed = elapsed + (t1-t0)

       deallocate(A, b)

       m = m+1

       goto 10
     end if

     deallocate(work)

     write(*, '(a,i2,a,d10.5)'), ' N = 2^', i, ', average time elapsed for LLS computation = ', elapsed/m
   end do

   write ( *, '(a)' ) ' '
end



