! Fortran/LAPACK version of some MATLAB codes from 
! Spectral Methods in MATLAB by L.N. Trefethen
! 
! David Knezevic, October 2006.

program main
  implicit none

  call p13
  call p14
  call p15

end

subroutine p13
  ! Fortran version of p13.m from Spectral Methods in MATLAB by L.N. Trefethen
  ! Solve linear BVP u_xx = exp(4x), u(-1)=u(1)=0

  implicit none
  integer, parameter :: N = 20
  integer info      ! factorisation/solving error flag
  integer ipiv(N-1) ! row interchanges
  integer i, j

  real (kind = 8) D(0:N, 0:N), D2temp(0:N, 0:N), D2(1:N-1,1:N-1), x(0:N), f(1:N-1)
  real (kind = 8) u(1:N-1), diff(1:N-1)
  real (kind = 8) dvec_norm_inf

  write(*, * ), '-------------------------------------------------------'
  write(*, * )
  write(*, * ), 'Fortran version of p13 from Spectral Methods in MATLAB:'
  write(*, * )
  write( *, *)

  call cheb(D, x, N)
  D2temp = matmul(D,D) ! Intrinsic Fortran function
!   call dgemm( 'n', 'n', N+1, N+1, N+1, 1.0d0, D,  N+1, D, N+1, 0.0d0, D2temp, N+1 ) ! BLAS subroutine for matrix multiplication

  D2(1:N-1,1:N-1) = D2temp(1:N-1,1:N-1)   ! We aren't interested in outer shell of the matrix

   do i = 1,N-1
     f(i) = exp(4.*x(i));
   end do

  call dgetrf ( N-1, N-1, D2, N-1, ipiv, info )   !  LU factorise the matrix.

  if ( info /= 0 ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a,i8)' ) '  Matrix is singular, INFO = ', info
    return
  end if


  call dgetrs ( 'n', N-1, 1, D2, N-1, ipiv, f, N-1, info )   ! Solve the linear system
  u(1:N-1) = f(1:N-1)

  if ( info /= 0 ) then
    write ( *, '(a)' ) ' '
    write ( *, '(a,i8)' ) '  Solution procedure failed, INFO = ', info
    return
  end if

  do i = 1,N-1
    diff(i) = ( exp(4.*x(i)) - sinh(4.)*x(i) - cosh(4.) ) / 16. - u(i) ! Difference between computed and exact sol at cheb pts
  end do

  write( *, *),           'exact solution = exp(4.*x(i)) - sinh(4.)*x(i) - cosh(4.)) / 16.'
  write( *, *)
  write( *, '(a,i3,a,d10.4)' ) ' N = ', N, ', l-infty norm of error at cheb pts = ', dvec_norm_inf(N-1,diff) ! Print out l-infty norm of error
  write(*, * )
  write(*, * ), '-------------------------------------------------------'
  write(*, * )

end

subroutine p14
  ! Fortran version of p14.m from Spectral Methods in MATLAB by L.N. Trefethen
  ! Solve nonlinear BVP u_xx = exp(u), u(-1)=u(1)=0

  implicit none
  integer, parameter :: N = 16
  integer info      ! factorisation/solving error flag
  integer ipiv(N-1) ! row interchanges
  integer i, it

  real (kind = 8) D(0:N, 0:N), D2temp(0:N, 0:N), D2(1:N-1,1:N-1), x(0:N), f(1:N-1)
  real (kind = 8) u(1:N-1), rhs(1:N-1), diff(1:N-1)
  real (kind = 8) dvec_norm_2
  real (kind = 8) change

  write(*, * ), 'Fortran version of p14 from Spectral Methods in MATLAB: '
  write(*, * )

  call cheb(D, x, N)
  D2temp = matmul(D,D)
!   call dgemm( 'n', 'n', N+1, N+1, N+1, 1.0d0, D,  N+1, D, N+1, 0.0d0, D2temp, N+1 ) ! BLAS subroutine for matrix multiplication

  D2(1:N-1,1:N-1) = D2temp(1:N-1,1:N-1)   ! We aren't interested in outer shell of the matrix

  u(1:N-1) = 0.0d0 ! zero u

  change = 1.
  it = 0

  call dgetrf ( N-1, N-1, D2, N-1, ipiv, info )                     ! LU factorise the matrix.


10 if (change .gt. 1.e-15) then                                     ! fixed point iteration

      do i = 1,N-1
        rhs(i) = exp(u(i))
      end do

      call dgetrs ( 'n', N-1, 1, D2, N-1, ipiv, rhs, N-1, info )    ! Solve the linear system

      do i = 1,N-1
        diff(i) = rhs(i) - u(i)                                     ! difference between consecutive iterations
      end do

      u(1:N-1) = rhs(1:N-1)

      change = dvec_norm_2(N-1,diff)
      it = it + 1

      write(*,'(a,i2,a,d10.5)'), ' iteration = ', it, ', change = ', change

      goto 10
   endif

  write(*, * )
  write(*, '(a,i3,a,i2,a,f10.7)'), ' N = ', N, ', no. of steps = ', it, ', u(0) = ', u(N/2.)
  write(*, * )
  write(*, * ), '-------------------------------------------------------'
  write(*, * )

end

subroutine p15
  ! Fortran version of p15.m from Spectral Methods in MATLAB by L.N. Trefethen
  ! Solve eigenvalue BVP u_xx = lamba*u, u(-1) = u(1) = 0

  implicit none
  integer, parameter :: N = 36
  integer i, info
  integer, parameter :: lwork = 4*(N-1)
  integer Y(1:N-1)

  real (kind = 8), parameter :: pi = 3.14159

  real (kind = 8) D(0:N, 0:N), x(0:N), D2temp(0:N, 0:N), D2(1:N-1,1:N-1), vl(1:N-1,1:N-1), vr(1:N-1,1:N-1)
  real (kind = 8) wr(1:N-1), wi(1:N-1), work(lwork)

  real (kind = 4) wr_single(1:N-1)

  write(*, * ), 'Fortran version of p15 from Spectral Methods in MATLAB: '
  write(*, * )

  call cheb(D, x, N)
  D2temp = matmul(D,D)

  D2(1:N-1,1:N-1) = D2temp(1:N-1,1:N-1)   ! We aren't interested in outer shell of the matrix

  ! Use sgeev to compute the left eigenvalues for a general matrix (D2 is non-symmetric)
  ! for these sorting utilities, we need to use single precision
  call dgeev('v', 'n',N-1, D2, N-1, wr, wi, vl, N-1, vr, N-1, work, lwork, info)

  wr_single(1:N-1) = wr(1:N-1) ! Sorting routine requires single precision reals

  call ssort(wr_single, Y, N-1, -1)

  do i=5,30,5
    write(*,'(a,i2,a,f11.5,a)'), ' eig ',i, ' = ', wr_single(i)*4.*pi**(-2.), '*4/pi^2'
  end do

  write(*, * )
  write(*, * ), '-------------------------------------------------------'
  write(*, * )

end


subroutine cheb ( D, x, N )
  ! Fortran version of cheb.m from Spectral Methods in MATLAB by L.N. Trefethen
  implicit none

  real (kind = 8) D(0:N, 0:N)
  real (kind = 8) x(0:N)
  integer N

  integer i, j
  real (kind = 8), parameter :: pi = 3.14159

  do i = 0,N
    x(i) = cos(pi*i/N)
  end do


  D(0,0) = (2.*N*N+1.)/6.
  D(N,0) = -0.5*((-1.)**N)
  D(0,N) = -D(N,0)
  D(N,N) = -D(0,0)

  do i = 1, N-1
    D(0,i) = 2.*(-1.)**i/(1-x(i))
    D(N,i) = -2.*((-1.)**(N+i))/(1+x(i))

    D(i,0) = -0.5*(-1.)**i/(1-x(i))
    D(i,N) = 0.5*(-1.)**(N+i)/(1+x(i))
  end do

  do i = 1, N-1
    do j = 1, N-1
      if (i == j) then
        D(i,j) = -x(j)/2./(1.-x(j)**2.)
      else
        D(i,j) = (-1)**(i+j)/(x(i)-x(j))
      endif
    end do
  end do

end

function dvec_norm_2 (n, u)

  implicit none
  integer n, i

  real (kind = 8) u(n)
  real (kind = 8) dvec_norm_2

  do i = 1,n
    dvec_norm_2 = dvec_norm_2 + u(i)**2
  end do

  dvec_norm_2 = sqrt(dvec_norm_2)

end

function dvec_norm_inf (n, u)

  implicit none
  integer n, i

  real (kind = 8) u(n), dvec_norm_inf

  dvec_norm_inf = 0.0d0

  do i = 1,n
    dvec_norm_inf = max (dvec_norm_inf, abs(u(i)))
  end do

end
